#include <windows.h>
#include <stdarg.h>
#include <string.h>
#include <shellapi.h>
#include <errno.h>
#include <stdlib.h>

// Standard Window Colors
#define COLOR_BLACK			PALETTERGB(0, 0, 0)
#define COLOR_RED				PALETTERGB(191, 0, 0)
#define COLOR_GREEN			PALETTERGB(0, 191, 0)
#define COLOR_BROWN			PALETTERGB(171, 87, 0)
#define COLOR_YELLOW			PALETTERGB(128, 128, 0)
#define COLOR_BLUE			PALETTERGB(0, 0, 191)
#define COLOR_MAGENTA		PALETTERGB(191, 0, 191)
#define COLOR_CYAN			PALETTERGB(0, 191, 191)
#define COLOR_WHITE			PALETTERGB(192, 192, 192)
#define COLOR_LIGHTBLACK	PALETTERGB(128, 128, 128)
#define COLOR_LIGHTRED		PALETTERGB(255, 0, 0)
#define COLOR_LIGHTGREEN	PALETTERGB(0, 255, 0)
#define COLOR_LIGHTYELLOW	PALETTERGB(255, 255, 0)
#define COLOR_LIGHTBLUE		PALETTERGB(0, 0, 255)
#define COLOR_LIGHTMAGENTA	PALETTERGB(255, 0, 255)
#define COLOR_LIGHTCYAN		PALETTERGB(0, 255, 255)
#define COLOR_LIGHTWHITE	PALETTERGB(255, 255, 255)

// Compatability Constants
#define BLACK					0
#define BLUE					1
#define GREEN					2
#define CYAN					3
#define RED						4
#define MAGENTA				5
#define BROWN					6
#define LIGHTGRAY				7
#define DARKGRAY				8
#define LIGHTBLUE				9
#define LIGHTGREEN			10
#define LIGHTCYAN				11
#define LIGHTRED				12
#define LIGHTMAGENTA			13
#define YELLOW					14
#define WHITE					15
#define BLINK					128

// WM_PERFORMFUNCTION notifications
#define SCR_PRINTF			1
#define SCR_SETPOS			2
#define SCR_SCANF				3
#define SCR_READBYTE			4
#define SCR_SETFGCOLOR		5
#define SCR_SETBGCOLOR		6
#define SCR_GETFGCOLOR		7
#define SCR_GETBGCOLOR		8
#define SCR_SETCLOSE			9
#define SCR_KBHIT				10
#define SCR_PRINTTOGGLE		11
#define SCR_PRINTSCREEN		12
#define SCR_PRINTSETUP		13
#define SCR_ABOUTTEXT		14
#define SCR_DISPLAYABOUT	15
#define SCR_APPCLOSE			16
#define SCR_GETFONTMULTIPLIERS 17
#define SCR_COLORCHOOSER   18
#define SCR_BGFGCOLORS		19
#define SCR_PLUGININIT		20
#define SCR_GETEXTENTS		21
#define SCR_SETEXTENTS		22
#define SCR_GETPOS			23
#define SCR_PAINT				24
#define SCR_COPY				25
#define SCR_PASTE				26
#define SCR_CANPASTE			27
#define SCR_GETSELECTION	28
#define SCR_SETSELECTION	29
#define SCR_SELECTALL		30
#define SCR_GETSELLENGTH   31
#define SCR_GETSELTEXT     32
#define SCR_GETCLOSE			33
#define SCR_SETINPUTLEN		34
#define SCR_GETINPUTLEN		35
#define SCR_FILEDIALOG		36
#define SCR_GETPLUGIN      37
#define SCR_CLEARINPUT		38
#define SCR_CLEARLINE		39
#define SCR_SETWINDOWTEXT  40
#define SCR_GETWINDOWTEXT  41
#define SCR_SETTABSPACE		42
#define SCR_GETTABSPACE		43
#define SCR_CLEARSCR			44

// Notification message
UINT SCR_PERFORMFUNCTION;

// Generic output string
HGLOBAL houtputStr;

// Instance variable
HINSTANCE hInstance;

// Handle to the output window
HWND hWnd;

// Locks (only 1 program may use us at a time; sorry..)
BOOL locked;

// putsNL Mode
BOOL putsNL;

// Exit flag
BOOL exitProgramOnClose;

// Prototype for vsscanf
int     _RTLENTRY _EXPFUNC vsscanf(const char * __buffer, const char * __format,
								void * __arglist);

#define _EXPORT huge
BOOL FAR PASCAL _export scrInitialize(LPSTR windowName, int x, int y, int rows, int columns, BOOL closeFlag);
void FAR PASCAL _export scrUninitialize(void);

void FAR PASCAL _export clreol(void);
void FAR PASCAL _export clrscr(void);
void FAR PASCAL _export gotoxy(int x, int y);
int FAR PASCAL _export wherex(void);
int FAR PASCAL _export wherey(void);
int FAR PASCAL _export kbhit(void);
int FAR PASCAL _export printf(LPCSTR format, ...);
int FAR PASCAL _export putch(int c);
int FAR PASCAL _export putchar(int c);
int FAR PASCAL _export puts(LPCSTR s);
void FAR PASCAL _export puti(int i);
void FAR PASCAL _export putpascalstring(LPCSTR str);
void FAR PASCAL _export textcolor(int newcolor);
void FAR PASCAL _export textbackground(int newcolor);
void FAR PASCAL _export setAboutText(LPSTR text);
void FAR PASCAL _export displayAbout(void);
void FAR PASCAL _export flushinput(void);
int FAR PASCAL _export fgetchar(void);
int FAR PASCAL _export getch(void);
int FAR PASCAL _export getchar(void);
int FAR PASCAL _export getche(void);
LPSTR FAR PASCAL _export gets(LPSTR s);
LPSTR FAR PASCAL _export cgets(LPSTR s);
void FAR PASCAL _export perror(LPCSTR s);
int FAR PASCAL _export vprintf(LPCSTR format, va_list arglist);
int FAR PASCAL _export vscanf(LPCSTR format, va_list arglist);
int FAR PASCAL _export scanf(LPCSTR format, ...);
void FAR PASCAL _export setCompatibilityMode(BOOL standardC);
void FAR PASCAL _export setWindowExitMode(BOOL closeProgram);
//int FAR PASCAL _export movetext(int left, int top, int right, int bottom, int destleft, int desttop);
//int FAR PASCAL _export ungetch(int ch);

COLORREF colorValue(int newcolor);
BOOL invalidWindowHandler(void);

BOOL FAR PASCAL _export scrInitialize(LPSTR windowName, int x, int y, int rows, int columns, BOOL closeFlag)
{
  // Variables
  HINSTANCE hWindowInst;
  LPSTR lpszCmdLine;
  DWORD time;
  HGLOBAL hGlobal;
  POINT pt;

  // Check the parameters
  if ((rows < 1) || (rows > 200) || (columns < 1) || (columns > 200))
	 return FALSE;

  // Check locks (return FALSE if this DLL is locked)
  if (locked)
	 return FALSE;

  // Allocate the lpszCmdLine string:    SCRWINDO.EXE + window name  +  $  + x & y + rows & columns + closeFlag
  hGlobal = GlobalAlloc(GHND, sizeof(BYTE) * (1 + lstrlen(windowName) + 1 + 2 * 6 + 2 * 4 + 2));
  lpszCmdLine = (LPSTR) GlobalLock(hGlobal);

  // Assign the values to lpszCmdLine
  lpszCmdLine[0] = '\0';
  wsprintf(lpszCmdLine, " %s$%i %i %i %i %i", windowName, x, y, rows, columns, closeFlag);

  // Launch the SCRWINDO.EXE application (if it could not be loaded, return an error
  if ((hWindowInst = ShellExecute(NULL, NULL, "SCRWINDO.EXE", lpszCmdLine, NULL, SW_SHOWNORMAL)) < HINSTANCE_ERROR)
	 return FALSE;

  // Get the window handle; if we can't in 6.5 seconds, return FALSE
  pt.x = x + 1;
  pt.y = y + 1;
  time = GetTickCount();
  while (((HINSTANCE) GetWindowWord((hWnd = WindowFromPoint(pt)), GWW_HINSTANCE) != hWindowInst) && (GetTickCount() <= time + 6500))
	 Yield();

  // Free the lpszCmdLine memory object
  GlobalUnlock(hGlobal);
  GlobalFree(hGlobal);

  // If the hWnd that we got is not the correct one, return FALSE
  if ((HINSTANCE) GetWindowWord(hWnd, GWW_HINSTANCE) != hWindowInst)
	 return FALSE;

  // Allocate the generic outputStr variable to size of screen or 255, which ever is larger
  houtputStr = GlobalAlloc(GHND | GMEM_SHARE, sizeof(BYTE) * ((rows * columns > 256)? (rows * columns): 256));

  // Set default putsNL mode, which is TRUE
  putsNL = TRUE;

  // Lock DLL and return TRUE
  locked = TRUE;
  return TRUE;
}

// DLL Entry Point
#pragma argsused
int FAR PASCAL LibMain(HINSTANCE hinst, WORD wDataSeg, WORD wHeapSize, LPSTR lpszCmdLine)
{
  // Standard DLL stuff
  if (wHeapSize > 0)
	 UnlockData(0);

  // Initialize hInstance (used when we create the window)
  hInstance = hinst;

  // Initialize the message
  SCR_PERFORMFUNCTION = RegisterWindowMessage("scr_performfunctionbystevenlawrance");

  // Initialize lock to FALSE
  locked = FALSE;

  // Initialize close flag to TRUE
  exitProgramOnClose = TRUE;
  return TRUE;
}

int FAR PASCAL _export scanf(LPCSTR format, ...)
{
  // Call vscanf with the arguments
  va_list arglist;
  int returnInfo;

  va_start(arglist, format);
  returnInfo = vscanf(format, arglist);
  va_end(arglist);

  return returnInfo;
}

int FAR PASCAL _export vscanf(LPCSTR format, va_list arglist)
{
  // Get the argument list (and don't argue! :-)
  HGLOBAL hinputStr = GlobalAlloc(GHND | GMEM_SHARE, SendMessage(hWnd, SCR_PERFORMFUNCTION, SCR_GETINPUTLEN, NULL));
  LPSTR inputStr = GlobalLock(hinputStr);
  DWORD returnInfo;

  // Check to see if the window is still running
  if (!IsWindow(hWnd))
	 return invalidWindowHandler();


  // Get the data from the input window
  SendMessage(hWnd, SCR_PERFORMFUNCTION, SCR_SCANF, (LPARAM) inputStr);

  // Convert the string into parameters
  returnInfo = vsscanf(inputStr, format, arglist);

  // Unlock the memory and free it
  GlobalUnlock(hinputStr);
  GlobalFree(hinputStr);

  // Advance to the next line
  SendMessage(hWnd, SCR_PERFORMFUNCTION, SCR_PRINTF, (LPARAM) "\n");

  // Return the number of bytes outputted
  return (int) returnInfo;
}


int FAR PASCAL _export printf(LPCSTR format, ...)
{
  // Call vprintf with the arguments
  va_list arglist;
  int returnInfo;

  va_start(arglist, format);
  returnInfo = vprintf(format, arglist);
  va_end(arglist);

  return returnInfo;
}

int FAR PASCAL _export vprintf(LPCSTR format, va_list arglist)
{
  // Get the argument list (and don't argue! :-)
  LPSTR outputStr;
  DWORD returnInfo;

  // Lock the memory and "clear" it
  if ((outputStr = (LPSTR) GlobalLock(houtputStr)) == NULL)
	 return FALSE;
  outputStr[0] = '\0';

  // Convert the parameters into a string
  wvsprintf(outputStr, format, arglist);

  // Check to see if the window is still running
  if (!IsWindow(hWnd))
	 return invalidWindowHandler();

  // Send printf to the window, and set our new coordinates
  SendMessage(hWnd, SCR_PERFORMFUNCTION, SCR_PRINTF, (LPARAM) outputStr);

  // Get the number of bytes outputted
  returnInfo = lstrlen(outputStr);

  // Unlock the memory
  GlobalUnlock(houtputStr);

  // Return the number of bytes outputted
  return (int) returnInfo;
}

void FAR PASCAL _export clreol(void)
{
  // Check to see if the window is still running
  if (!IsWindow(hWnd))
  {
	 invalidWindowHandler();
	 return;
  }

  SendMessage(hWnd, SCR_PERFORMFUNCTION, SCR_CLEARLINE, 0);
}

void FAR PASCAL _export clrscr(void)
{
  // Check to see if the window is still running
  if (!IsWindow(hWnd))
  {
	 invalidWindowHandler();
	 return;
  }

  SendMessage(hWnd, SCR_PERFORMFUNCTION, SCR_CLEARSCR, 0);
}

void FAR PASCAL _export gotoxy(int x, int y)
{
  // Check to see if the window is still running
  if (!IsWindow(hWnd))
  {
	 invalidWindowHandler();
	 return;
  }

  SendMessage(hWnd, SCR_PERFORMFUNCTION, SCR_SETPOS, MAKELONG(y - 1, x - 1));
}

void FAR PASCAL _export displayAbout(void)
{
  // Check to see if the window is still running
  if (!IsWindow(hWnd))
  {
	 invalidWindowHandler();
	 return;
  }

  SendMessage(hWnd, SCR_PERFORMFUNCTION, SCR_DISPLAYABOUT, 0);
}

int FAR PASCAL _export wherex(void)
{
  DWORD pos;

  // Check to see if the window is still running
  if (!IsWindow(hWnd))
	 return invalidWindowHandler();

  pos = SendMessage(hWnd, SCR_PERFORMFUNCTION, SCR_GETPOS, 0);
  return (HIWORD(pos) + 1);
}

int FAR PASCAL _export wherey(void)
{
  DWORD pos;

  // Check to see if the window is still running
  if (!IsWindow(hWnd))
	 return invalidWindowHandler();

  pos = SendMessage(hWnd, SCR_PERFORMFUNCTION, SCR_GETPOS, 0);
  return (LOWORD(pos) + 1);
}

void FAR PASCAL _export puti(int i)
{
  char outText[6];

  // Check to see if the window is still running
  if (!IsWindow(hWnd))
  {
	 invalidWindowHandler();
	 return;
  }

  wsprintf(outText, "%i", i);
  SendMessage(hWnd, SCR_PERFORMFUNCTION, SCR_PRINTF, (LPARAM)outText);
}

void FAR PASCAL _export putpascalstring(LPCSTR str)
{
  LPSTR newStr;

  // Check to see if the window is still running
  if (!IsWindow(hWnd))
  {
	 invalidWindowHandler();
	 return;
  }

  if (str[0] > 0)
  {
	 HGLOBAL hnewStr;
	 UINT i;
	 hnewStr = GlobalAlloc(GHND, sizeof(BYTE) * str[0] + 1);
	 newStr = GlobalLock(hnewStr);
	 for (i = 1; i <= str[0]; i++)
		newStr[i - 1] = str[i];
	 newStr[str[0]] = '\0';
	 SendMessage(hWnd, SCR_PERFORMFUNCTION, SCR_PRINTF, (LPARAM)newStr);
	 GlobalUnlock(hnewStr);
	 GlobalFree(hnewStr);
  }
}

int FAR PASCAL _export kbhit(void)
{
  // Check to see if the window is still running
  if (!IsWindow(hWnd))
	 return invalidWindowHandler();

  Yield();
  return (int) SendMessage(hWnd, SCR_PERFORMFUNCTION, SCR_KBHIT, 0);
}

int FAR PASCAL _export putch(int c)
{
  char outText[2];

  // Check to see if the window is still running
  if (!IsWindow(hWnd))
	 return invalidWindowHandler();

  outText[0] = c;
  outText[1] = '\0';
  SendMessage(hWnd, SCR_PERFORMFUNCTION, SCR_PRINTF, (LPARAM)outText);
  return 1;
}

int FAR PASCAL _export putchar(int c)
{
  char outText[2];

  // Check to see if the window is still running
  if (!IsWindow(hWnd))
	 return invalidWindowHandler();

  outText[0] = c;
  outText[1] = '\0';
  SendMessage(hWnd, SCR_PERFORMFUNCTION, SCR_PRINTF, (LPARAM) outText);
  return 1;
}

int FAR PASCAL _export puts(LPCSTR s)
{
  // Check to see if the window is still running
  if (!IsWindow(hWnd))
	 return invalidWindowHandler();

  // Write the string
  SendMessage(hWnd, SCR_PERFORMFUNCTION, SCR_PRINTF, (LPARAM) s);

  // Write a newline if putsNL is on
  if (putsNL)
	 SendMessage(hWnd, SCR_PERFORMFUNCTION, SCR_PRINTF, (LPARAM) "\n");
  return lstrlen(s);
}

void FAR PASCAL _export setCompatibilityMode(BOOL standardC)
{
  // For now, putsNL is the only thing we must set from standardC
  putsNL = standardC;
}

void FAR PASCAL _export textcolor(int newcolor)
{
  // Check to see if the window is still running
  if (!IsWindow(hWnd))
  {
	 invalidWindowHandler();
	 return;
  }

  SendMessage(hWnd, SCR_PERFORMFUNCTION, SCR_SETFGCOLOR, colorValue(newcolor));
}

void FAR PASCAL _export textbackground(int newcolor)
{
  // Check to see if the window is still running
  if (!IsWindow(hWnd))
  {
	 invalidWindowHandler();
	 return;
  }

  SendMessage(hWnd, SCR_PERFORMFUNCTION, SCR_SETBGCOLOR, colorValue(newcolor));
}

void FAR PASCAL _export flushinput(void)
{
  // Check to see if the window is still running
  if (!IsWindow(hWnd))
  {
	 invalidWindowHandler();
	 return;
  }

  // Flush the input buffer
  SendMessage(hWnd, SCR_PERFORMFUNCTION, SCR_CLEARINPUT, NULL);
}

void FAR PASCAL _export setAboutText(LPSTR text)
{
  // Check to see if the window is still running
  if (!IsWindow(hWnd))
  {
	 invalidWindowHandler();
	 return;
  }

  SendMessage(hWnd, SCR_PERFORMFUNCTION, SCR_ABOUTTEXT, (LPARAM) text);
}

void FAR PASCAL _export scrUninitialize(void)
{
  // Check to see if the window is still running
  if (!IsWindow(hWnd))
  {
	 invalidWindowHandler();
	 return;
  }

  // Deallocate output buffer and notify the window
  GlobalFree(houtputStr);
  SendMessage(hWnd, SCR_PERFORMFUNCTION, SCR_APPCLOSE, NULL);
}

int FAR PASCAL _export getch(void)
{
  // Check to see if the window is still running
  if (!IsWindow(hWnd))
	 return invalidWindowHandler();

  // Get the next character without the [Enter] requirement
  return (int) SendMessage(hWnd, SCR_PERFORMFUNCTION, SCR_READBYTE, FALSE);
}

int FAR PASCAL _export getche(void)
{
  // Get character and echo it to the screen
  int ch[2] = { '\0', '\0' };

  // Check to see if the window is still running
  if (!IsWindow(hWnd))
	 return invalidWindowHandler();

  ch[0] = (int) SendMessage(hWnd, SCR_PERFORMFUNCTION, SCR_READBYTE, FALSE);
  SendMessage(hWnd, SCR_PERFORMFUNCTION, SCR_PRINTF, (LPARAM) ch);
  return ch[0];
}

int FAR PASCAL _export fgetchar(void)
{
  // Get the next character with the [Enter] requirement
  int returnInfo = (int) SendMessage(hWnd, SCR_PERFORMFUNCTION, SCR_READBYTE, TRUE);

  // Check to see if the window is still running
  if (!IsWindow(hWnd))
	 return invalidWindowHandler();

  // Advance to the next line
  SendMessage(hWnd, SCR_PERFORMFUNCTION, SCR_PRINTF, (LPARAM) "\n");
  return returnInfo;
}

int FAR PASCAL _export getchar(void)
{
  // Get the next character with the [Enter] requirement
  int returnInfo = (int) SendMessage(hWnd, SCR_PERFORMFUNCTION, SCR_READBYTE, TRUE);

  // Check to see if the window is still running
  if (!IsWindow(hWnd))
	 return invalidWindowHandler();

  // Advance to the next line
  SendMessage(hWnd, SCR_PERFORMFUNCTION, SCR_PRINTF, (LPARAM) "\n");
  return returnInfo;
}

LPSTR FAR PASCAL _export gets(LPSTR s)
{
  // Check to see if the window is still running
  if (!IsWindow(hWnd))
	 return (LPSTR) invalidWindowHandler();

  // Read the string
  SendMessage(hWnd, SCR_PERFORMFUNCTION, SCR_SCANF, (LPARAM) s);
  return s;
}

LPSTR FAR PASCAL _export cgets(LPSTR s)
{
  // Check to see if the window is still running
  if (!IsWindow(hWnd))
	 return (LPSTR) invalidWindowHandler();

  // Read the string
  SendMessage(hWnd, SCR_PERFORMFUNCTION, SCR_SCANF, (LPARAM) s);
  return s;
}

void FAR PASCAL _export perror(LPCSTR s)
{
  // Print an error message
  LPSTR outputStr;

  // Check to see if the window is still running
  if (!IsWindow(hWnd))
  {
	 invalidWindowHandler();
	 return;
  }

  if ((outputStr = (LPSTR) GlobalLock(houtputStr)) == NULL)
	 return;
  outputStr[0] = '\0';

  // Format the string and print the result
  wsprintf(outputStr, "%s: Error %i\n", s, errno);
  SendMessage(hWnd, SCR_PERFORMFUNCTION, SCR_PRINTF, (LPARAM) outputStr);
  GlobalUnlock(houtputStr);
}

void FAR PASCAL _export setWindowExitMode(BOOL closeProgram)
{
  // Set the behavior when the user closes the window (default = close program too)
  exitProgramOnClose = closeProgram;
}

BOOL invalidWindowHandler(void)
{
  // The window is not present; exit the running application that we're attached to
  // if the flag is set to do so
  if (exitProgramOnClose)
	 exit(EXIT_SUCCESS);

  // If not, return FALSE to the program that called us
  return FALSE;
}

COLORREF colorValue(int newcolor)
{
  COLORREF color;
  switch (newcolor)
  {
	 case BLACK : color = COLOR_BLACK; break;
	 case RED : color = COLOR_RED; break;
	 case GREEN : color = COLOR_GREEN; break;
	 case BROWN : color = COLOR_BROWN; break;
	 case YELLOW : color = COLOR_YELLOW; break;
	 case BLUE : color = COLOR_BLUE; break;
	 case MAGENTA : color = COLOR_MAGENTA; break;
	 case CYAN : color = COLOR_CYAN; break;
	 case LIGHTGRAY : color = COLOR_WHITE; break;
	 case DARKGRAY : color = COLOR_LIGHTBLACK; break;
	 case LIGHTRED : color = COLOR_LIGHTRED; break;
	 case LIGHTGREEN : color = COLOR_LIGHTGREEN; break;
	 case LIGHTBLUE : color = COLOR_LIGHTBLUE; break;
	 case LIGHTMAGENTA : color = COLOR_LIGHTMAGENTA; break;
	 case LIGHTCYAN : color = COLOR_LIGHTCYAN; break;
	 case WHITE : color = COLOR_LIGHTWHITE; break;
  }
  return color;
}

