// ScreenWindow Source code
// By Steven Lawrance

#include <windows.h>
#include <stdlib.h>
#include <string.h>
#include <ctl3d.h>
#include "scrwindo.h"

// Class name (as in window class)
#define SCR_CLASSNAME      "slscreen385473"

// Default about text
#define DEFAULTABOUTTEXT   "Windows Program"

// Default tab space
#define SCR_TABSPACEDEFAULT 5

// Font multipliers (with Terminal fixed-space font)
#define SCR_X              8
#define SCR_Y					12

// Identification number of the timers
#define STARTUPID				610
#define CARETID				615

// WM_PERFORMFUNCTION notifications
#define SCR_PRINTF			1
#define SCR_SETPOS			2
#define SCR_SCANF				3
#define SCR_READBYTE			4
#define SCR_SETFGCOLOR		5
#define SCR_SETBGCOLOR		6
#define SCR_GETFGCOLOR		7
#define SCR_GETBGCOLOR		8
#define SCR_SETCLOSE			9
#define SCR_KBHIT				10
#define SCR_PRINTTOGGLE		11
#define SCR_PRINTSCREEN		12
#define SCR_PRINTSETUP		13
#define SCR_ABOUTTEXT		14
#define SCR_DISPLAYABOUT	15
#define SCR_APPCLOSE			16
#define SCR_GETFONTMULTIPLIERS 17
#define SCR_COLORCHOOSER   18
#define SCR_BGFGCOLORS		19
#define SCR_PLUGININIT		20
#define SCR_GETEXTENTS		21
#define SCR_SETEXTENTS		22
#define SCR_GETPOS			23
#define SCR_PAINT				24
#define SCR_COPY				25
#define SCR_PASTE				26
#define SCR_CANPASTE			27
#define SCR_GETSELECTION	28
#define SCR_SETSELECTION	29
#define SCR_SELECTALL		30
#define SCR_GETSELLENGTH   31
#define SCR_GETSELTEXT     32
#define SCR_GETCLOSE			33
#define SCR_SETINPUTLEN		34
#define SCR_GETINPUTLEN		35
#define SCR_FILEDIALOG		36
#define SCR_GETPLUGIN      37
#define SCR_CLEARINPUT		38
#define SCR_CLEARLINE		39
#define SCR_SETWINDOWTEXT  40
#define SCR_GETWINDOWTEXT  41
#define SCR_SETTABSPACE		42
#define SCR_GETTABSPACE		43
#define SCR_CLEARSCR			44

// Standard Colors
#define COLOR_BLACK			PALETTERGB(0, 0, 0)
#define COLOR_RED				PALETTERGB(191, 0, 0)
#define COLOR_GREEN			PALETTERGB(0, 191, 0)
#define COLOR_BROWN			PALETTERGB(171, 87, 0)
#define COLOR_YELLOW			PALETTERGB(128, 128, 0)
#define COLOR_BLUE			PALETTERGB(0, 0, 191)
#define COLOR_MAGENTA		PALETTERGB(191, 0, 191)
#define COLOR_CYAN			PALETTERGB(0, 191, 191)
#define COLOR_WHITE			PALETTERGB(192, 192, 192)
#define COLOR_LIGHTBLACK	PALETTERGB(128, 128, 128)
#define COLOR_LIGHTRED		PALETTERGB(255, 0, 0)
#define COLOR_LIGHTGREEN	PALETTERGB(0, 255, 0)
#define COLOR_LIGHTYELLOW	PALETTERGB(255, 255, 0)
#define COLOR_LIGHTBLUE		PALETTERGB(0, 0, 255)
#define COLOR_LIGHTMAGENTA	PALETTERGB(255, 0, 255)
#define COLOR_LIGHTCYAN		PALETTERGB(0, 255, 255)
#define COLOR_LIGHTWHITE	PALETTERGB(255, 255, 255)

// Structure for the current color and every character's color
typedef struct
{
  COLORREF bgColor;
  COLORREF fgColor;
} COLORINFO;
typedef COLORINFO FAR*  LPCOLORINFO;

// Clipboard variables
typedef struct
{
  HWND hNext;
  int beginRow;
  int beginColumn;
  int endRow;
  int endColumn;
  int clickRow;
  int clickColumn;
  int x;
} CLIPINFO;
CLIPINFO clipboard;

// Background and foreground colors
COLORINFO curColor;

// Echo flag
BOOL echoChars;

// Input start location
UINT inRow;
UINT inColumn;

// Current location
UINT curRow;
UINT curColumn;

// Screen constraints
UINT numRows;
UINT numColumns;

// Handle to the menu of the window
HMENU hMenu;

// Tab space
UINT tabSpace;

// closeOnExit flag
BOOL closeOnExit;

// Notification message
UINT SCR_PERFORMFUNCTION;

// Instance variable
HINSTANCE hInstance;

// About Screen text (from program)
HGLOBAL haboutText;

// Caret Variables
BOOL caretStatus;
BOOL caretActive;
BOOL caretHide;

// Screen Buffer and their color values
LPSTR *scr;
LPCOLORINFO **color;

// Input buffer and length
HGLOBAL hscrIn;
LPSTR scrIn;
UINT scrInLength;

// Window handle
HWND hWnd;

// Global palette
HPALETTE hPalette;

// Global font
HFONT hFont;

// Keyboard shortcuts
HACCEL hAcc;

#define MAKEWORD(low, high) ((WORD)(((BYTE)(low)) | (((WORD)((BYTE)(high))) << 8)))

LRESULT CALLBACK _export MainWndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK _export AboutDlgProc(HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam);
HPALETTE makePalette(void);
UINT getSelectedTextLength(void);
BOOL getSelectedText(LPSTR str);
void SetCaretState(BOOL state, BOOL on);
void waitForEoln(void);
BOOL waitForChar(void);
int getLastCharPos(LPCSTR line);
BOOL isCharSelected(UINT r, UINT c);


int getValue(LPSTR lpszCmdLine, int *start)
{
  // Get the next integer in the command line
  char str[10];
  char i;
  int len = lstrlen(lpszCmdLine);
  memset(str, '\0', 10);

  // Go through the parameter list, stopping when we're at the end or have a space
  i = *start;
  while (i <= len - 1)
  {
	 if ((i - *start <= 8) && (lpszCmdLine[i] != ' '))
		str[i - *start] = lpszCmdLine[i];
	 else
		break;
	 i++;
  }
  *start = i + 1;

  // Return the integer version of the string if the string has a length;
  // otherwise, return -1
  if (lstrlen(str) > 0)
	 return atoi(str);
  else
	 return -1;
}


// Program Entry Point
#pragma argsused
int FAR PASCAL WinMain(HINSTANCE hinst, HINSTANCE hPrevInstance, LPSTR lpszCmdLine, int nCmdShow)
{
  HGLOBAL hGlobal, hwindowName;
  WNDCLASS wndclass;
  MSG msg;
  LPSTR windowName;
  LPSTR aboutText;
  int len;
  int x = CW_USEDEFAULT, y = CW_USEDEFAULT;
  int rows = 25, columns = 80;
  int i = 0, t;

  // Check parameters and assign values to their variables
  len = lstrlen(lpszCmdLine);
  hwindowName = GlobalAlloc(GHND, sizeof(BYTE) * len + 1);
  windowName = (LPSTR) GlobalLock(hwindowName);
  memset(windowName, '\0', sizeof(BYTE) * len + 1);
  while (i <= len - 1)
  {
	 if ((lpszCmdLine[i] == '$'))
		break;
	 i++;
  }
  if (i == 0)
	 lstrcpy(windowName, "ScreenWindow 1.00.96 alpha5");
  else
  {
	 // Get the window name
	 lstrcpyn(windowName, lpszCmdLine, i + 1);

	 // Get the other values; assign CW_USEDEFAULT if the parameter wasn't there
	 i++;
	 if ((x = getValue(lpszCmdLine, &i)) != -1)
	 {
		if ((y = getValue(lpszCmdLine, &i)) != -1)
		{
		  if ((rows = getValue(lpszCmdLine, &i)) != -1)
		  {
			 if ((columns = getValue(lpszCmdLine, &i)) != -1)
			 {
				if ((closeOnExit = getValue(lpszCmdLine, &i)) == -1)
				  closeOnExit = FALSE;
			 }
			 else
				columns = 80;
		  }
		  else
			 rows = 25;
		}
		else
		{
		  x = CW_USEDEFAULT;
		  y = CW_USEDEFAULT;
		}
	 }
	 else
	 {
		x = CW_USEDEFAULT;
		y = CW_USEDEFAULT;
	 }
  }

  // Initialize hInstance (used when we create the window)
  hInstance = hinst;

  // Set the current row and column to the top left
  curRow = 0;
  curColumn = 0;

  // Set the default colors to white on a black background
  curColor.bgColor = COLOR_BLACK;
  curColor.fgColor = COLOR_WHITE;

  // Define the class (automatically unregistered when DLL is unloaded :-)
  wndclass.lpszClassName = SCR_CLASSNAME;
  wndclass.lpfnWndProc   = (WNDPROC) MainWndProc;
  wndclass.cbClsExtra    = 0;
  wndclass.cbWndExtra    = 0;
  wndclass.hInstance     = hinst;
  wndclass.hIcon         = LoadIcon(NULL, IDI_APPLICATION);
  wndclass.hCursor       = LoadCursor(NULL, IDC_IBEAM);
  wndclass.hbrBackground = (HBRUSH) (COLOR_WINDOW + 1);
  wndclass.lpszMenuName  = NULL;
  wndclass.style         = CS_VREDRAW | CS_HREDRAW;

  // Register the class and return TRUE
  RegisterClass(&wndclass);

  // Check the parameters
  if ((rows < 1) || (rows > 200))
  {
	 // Assign the default value to the rows
	 rows = 25;
  }
  if ((columns < 1) || (columns > 200))
  {
	 // Assign the default value to the columns
	 columns = 80;
  }

  // Initialize the message
  SCR_PERFORMFUNCTION = RegisterWindowMessage("scr_performfunctionbystevenlawrance");

  // Assign the default input buffer length and allocate that space
  scrInLength = rows * columns + 1;
  hscrIn = GlobalAlloc(GHND, sizeof(BYTE) * scrInLength);
  scrIn = GlobalLock(hscrIn);
  scrIn[0] = '\0';

  // Allocate and set the default about dialog box text
  haboutText = GlobalAlloc(GHND, sizeof(BYTE) * (lstrlen(DEFAULTABOUTTEXT) + 1));
  aboutText = GlobalLock(haboutText);
  lstrcpy(aboutText, DEFAULTABOUTTEXT);
  GlobalUnlock(haboutText);

  // Allocate screen buffer to user's size, setting each character to a space in the process
  hGlobal = GlobalAlloc(GHND, sizeof(LPSTR) * rows); // allocate string pointers
  scr = (LPSTR *) GlobalLock(hGlobal);
  hGlobal = GlobalAlloc(GHND, sizeof(LPCOLORINFO **) * rows); // allocate COLORINFO pointers
  color = (LPCOLORINFO **) GlobalLock(hGlobal);
  for (i = 0; i <= rows - 1; i++)
  {
	 // Each row's string
	 hGlobal = GlobalAlloc(GHND, sizeof(BYTE) * (columns + 1)); // assign pointers
	 scr[i] = (LPSTR) GlobalLock(hGlobal);
	 memset(scr[i], ' ', sizeof(BYTE) * columns);
	 scr[i][columns] = '\0';

	 // Each row's color values
	 hGlobal = GlobalAlloc(GHND, sizeof(LPCOLORINFO *) * columns); // allocate COLORINFO pointers
	 color[i] = (LPCOLORINFO *) GlobalLock(hGlobal);

	 for (t = 0; t <= columns - 1; t++)
	 {
		hGlobal = GlobalAlloc(GHND, sizeof(COLORINFO)); // allocate COLORINFO pointers
		color[i][t] = (LPCOLORINFO) GlobalLock(hGlobal);
		color[i][t]->bgColor = curColor.bgColor;
		color[i][t]->fgColor = curColor.fgColor;
	 }
  }

  // Set the screen constraints
  numRows = rows;
  numColumns = columns;

  // Load the menu and its accelerators (key shortcuts)
  hMenu = LoadMenu(hInstance, "MAIN");
  hAcc = LoadAccelerators(hInstance, "MAIN");

  // Add 3D controls to our dialog boxes
  Ctl3dRegister(hInstance);
  Ctl3dAutoSubclass(hInstance);

  // Create window
  if ((hWnd = CreateWindow(SCR_CLASSNAME, windowName, WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX, x, y, 500, 500, NULL, hMenu, hinst, NULL)) == NULL)
	 return NULL;
  else
  {
	 // Free the windowName memory object
	 GlobalUnlock(hwindowName);
	 GlobalFree(hwindowName);

	 // Show window
	 ShowWindow(hWnd, nCmdShow);
	 UpdateWindow(hWnd);

	 // Message Loop
	 while (GetMessage(&msg, NULL, 0, 0))
	 {
		TranslateMessage(&msg);
		TranslateAccelerator(hWnd, hAcc, &msg);
		DispatchMessage(&msg);
	 }

	 // Clean up our objects, remove the class, and unregister the 3D dialog effects
	 DestroyMenu(hMenu);
	 DeleteObject(hFont);
	 UnregisterClass(SCR_CLASSNAME, hinst);
	 Ctl3dUnregister(hInstance);
	 return msg.wParam;
  }
}

#pragma argsused
LRESULT CALLBACK _export MainWndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
  switch (message)
  {
	 case WM_CREATE :
	 {
		// Retrieve the window and client rectangles
		RECT client, window;
		UINT wDiff, hDiff;
		GetClientRect(hWnd, &client);
		GetWindowRect(hWnd, &window);

		// Calculate the differences between the client's space and the window's space
		wDiff = (window.right - window.left) - client.right;
		hDiff = (window.bottom - window.top) - client.bottom;

		// Set the window's new size to account for the number of rows/columns
		SetWindowPos(hWnd, 0, 0, 0, numColumns * SCR_X + wDiff, numRows * SCR_Y + hDiff, SWP_NOZORDER | SWP_NOACTIVATE | SWP_NOMOVE);

		// Load the font into hFont
		hFont = CreateFont(12, 0, 0, 0, 0, 0, 0, 0, 255, 0, 0, 0, 0, "Terminal");

		// Create the palette and store it in hPalette
		hPalette = makePalette();

		// Link us into the clipboard chain
		clipboard.hNext = SetClipboardViewer(hWnd);

		// Check to see if the current clipboard data is CF_TEXT (something we can use)
		if (!IsClipboardFormatAvailable(CF_TEXT))
		  EnableMenuItem(hMenu, CM_EDITPASTE, MF_BYCOMMAND | MF_GRAYED);
		else
		  EnableMenuItem(hMenu, CM_EDITPASTE, MF_BYCOMMAND | MF_ENABLED);

		// Set the closeOnExit menu item
		if (closeOnExit)
		  CheckMenuItem(hMenu, CM_CLOSEONEXIT, MF_BYCOMMAND | MF_CHECKED);
		else
		  CheckMenuItem(hMenu, CM_CLOSEONEXIT, MF_BYCOMMAND | MF_UNCHECKED);

		// Set other clipboard information to defaults
		clipboard.beginRow = -1;
		clipboard.beginColumn = -1;
		clipboard.endRow = -1;
		clipboard.endColumn = -1;
		clipboard.clickRow = -1;
		clipboard.clickColumn = -1;
		clipboard.x = -1;

		// Set the echo flag to FALSE
		echoChars = FALSE;

		// Set the tab space to the default
		tabSpace = SCR_TABSPACEDEFAULT;

		// Set the input row/columns
		inRow = 0;
		inColumn = 0;

		// Perform other startup functions
		SetTimer(hWnd, STARTUPID, 1, NULL);
		return 0;

	 } /* WM_CREATE */

	 case WM_DRAWCLIPBOARD :
	 {
		// Check to see if the current clipboard data is CF_TEXT (something we can use)
		if (!IsClipboardFormatAvailable(CF_TEXT))
		  EnableMenuItem(hMenu, CM_EDITPASTE, MF_BYCOMMAND | MF_GRAYED);
		else
		  EnableMenuItem(hMenu, CM_EDITPASTE, MF_BYCOMMAND | MF_ENABLED);

		// Pass this message to the next clipboard window
		SendMessage(clipboard.hNext, WM_DRAWCLIPBOARD, wParam, lParam);
		return 0;
	 } /* WM_DRAWCLIPBOARD */

	 case WM_CHANGECBCHAIN :
	 {
		// Someone was removed from the clipboard chain; update our variables accordingly
		if ((HWND) wParam == clipboard.hNext)
		  clipboard.hNext = (HWND) LOWORD(lParam);

		// Pass this message to the next in the chain
		SendMessage(clipboard.hNext, WM_CHANGECBCHAIN, wParam, lParam);
		return 0;
	 }

	 case WM_RBUTTONUP :
	 {
		// Display the right-click menu, and destroy it when we're done
		POINT ptCurrent;
		HMENU hMenu;

		ptCurrent = MAKEPOINT(lParam);
		hMenu = CreatePopupMenu();
		AppendMenu(hMenu, MF_ENABLED, CM_EDITCOPY,      "&Copy");
		AppendMenu(hMenu, MF_ENABLED, CM_EDITPASTE,     "&Paste");
		AppendMenu(hMenu, MF_MENUBREAK, 0, NULL);
		AppendMenu(hMenu, MF_ENABLED, CM_EDITSELECTALL, "Select &All");
		ClientToScreen(hWnd, &ptCurrent);
		TrackPopupMenu(hMenu, TPM_LEFTALIGN | TPM_LEFTBUTTON | TPM_RIGHTBUTTON, ptCurrent.x, ptCurrent.y, 0, hWnd, NULL);
		DestroyMenu(hMenu);
		return 0;
	 }

	 case WM_LBUTTONDOWN :
	 {
		// The user is dragging text; store the coordinates in the clipboard structure
		clipboard.clickRow = HIWORD(lParam) / SCR_Y;
		clipboard.clickColumn = LOWORD(lParam) / SCR_X;
		clipboard.endRow = -1;
		clipboard.endColumn = -1;
		clipboard.x = LOWORD(lParam);
		return 0;
	 }

	 case WM_LBUTTONUP :
	 {
		// The user stopped selecting text; if no region was selected, end the selection
		// (with a pixel lenience of SCR_X / 2)
		clipboard.beginRow = clipboard.clickRow;
		clipboard.beginColumn = clipboard.clickColumn;
		clipboard.endRow = HIWORD(lParam) / SCR_Y;
		clipboard.endColumn = LOWORD(lParam) / SCR_X;
		if ((clipboard.endRow < clipboard.beginRow) || ((clipboard.endRow == clipboard.beginRow) && (clipboard.endColumn < clipboard.beginColumn)))
		{
		  // Swap the begin and the end
		  int swap;
		  swap = clipboard.endRow;
		  clipboard.endRow = clipboard.beginRow;
		  clipboard.beginRow = swap;
		  swap = clipboard.endColumn;
		  clipboard.endColumn = clipboard.beginColumn;
		  clipboard.beginColumn = swap;
		}
		clipboard.x -= LOWORD(lParam);
		if (clipboard.x < 0)
		  clipboard.x = -clipboard.x;
		if ((clipboard.endRow == clipboard.beginRow) && (clipboard.x < SCR_X / 2))
		{
		  // End the selection
		  clipboard.beginRow = -1;
		  clipboard.beginColumn = -1;
		  clipboard.endRow = -1;
		  clipboard.endColumn = -1;
        InvalidateRect(hWnd, NULL, FALSE);
		}
		clipboard.x = -1;
		return 0;
	 }

	 case WM_MOUSEMOVE :
	 {
		// Enlarge or shrink the selection, if there is one
		if (clipboard.x != -1)
		{
//		  RECT rect;
		  clipboard.beginRow = clipboard.clickRow;
		  clipboard.beginColumn = clipboard.clickColumn;
		  clipboard.endRow = HIWORD(lParam) / SCR_Y;
		  clipboard.endColumn = LOWORD(lParam) / SCR_X;
		  if ((clipboard.endRow < clipboard.beginRow) || ((clipboard.endRow == clipboard.beginRow) && (clipboard.endColumn < clipboard.beginColumn)))
		  {
			 // Swap the begin and the end
			 int swap;
			 swap = clipboard.endRow;
			 clipboard.endRow = clipboard.beginRow;
			 clipboard.beginRow = swap;
			 swap = clipboard.endColumn;
			 clipboard.endColumn = clipboard.beginColumn;
			 clipboard.beginColumn = swap;
		  }

		  // Calculate region to paint
/*		  rect.top = clipboard.beginRow * SCR_Y;
		  rect.left = 0;
		  rect.right = (numColumns - 1) * SCR_X;
		  rect.bottom = (clipboard.endRow + 1) * SCR_Y;*/

		  // Update the window
		  InvalidateRect(hWnd, NULL, FALSE);
		}
		return 0;
	 }

	 case WM_KILLFOCUS :
	 {
		// We're losing the focus :-( !!  Destroy the caret (sorry, Buggs Bunny!)
		SetCaretState(FALSE, FALSE);
		caretHide = TRUE;
		return 0;
	 }

	 case WM_SETFOCUS :
	 {
		// Set the caret status
		caretHide = FALSE;
		if (!caretStatus)
		  SetCaretState(!caretStatus, TRUE);

		// If there is a palette at this time, realize it to the system
		if (hPalette)
		{
		  HDC hDC = GetDC(hWnd);
		  HPALETTE hOldPal = SelectPalette(hDC, hPalette, FALSE);
		  RealizePalette(hDC);
		  SelectPalette(hDC, hOldPal, FALSE);
		  ReleaseDC(hWnd, hDC);
		}
		return 0;
	 }

	 case WM_TIMER :
	 {
		if (wParam == STARTUPID)
		{
		  // Now that the window is showing, let's recalculate the
		  // area in case the menu is larger than Windows had originally
		  // thought
		  RECT client, window;
		  UINT wDiff, hDiff;

		  // Destroy this timer
		  KillTimer(hWnd, STARTUPID);

		  // Retrieve the window and client rectangles
		  GetClientRect(hWnd, &client);
		  GetWindowRect(hWnd, &window);

		  // Calculate the differences between the client's space and the window's space
		  wDiff = (window.right - window.left) - client.right;
		  hDiff = (window.bottom - window.top) - client.bottom;

		  // Set the window's new size to account for the number of rows/columns
		  SetWindowPos(hWnd, 0, 0, 0, numColumns * SCR_X + wDiff, numRows * SCR_Y + hDiff, SWP_NOZORDER | SWP_NOACTIVATE | SWP_NOMOVE);

		  // Create a caret
		  caretHide = FALSE;
		  SetCaretState(FALSE, TRUE);

		  // Return 0
		  return 0;
		}
		else
		{
		  // Paint the caret, if it's active
		  if (caretActive)
			 SetCaretState(!caretStatus, TRUE);
		  return 0;
		}
	 }

	 case WM_ERASEBKGND :
	 {
		// Repaint everything
		InvalidateRect(hWnd, NULL, FALSE);
		return 0;
	 }

	 case WM_PAINT :
	 {
		// Repaint a portion (or all) of the screen
		HDC hDC = GetDC(hWnd);
		HFONT hOldFont = SelectObject(hDC, hFont);
		HPALETTE hOldPal = SelectPalette(hDC, hPalette, FALSE);
		RECT rect;
		COLORREF bgColor = curColor.bgColor, fgColor = curColor.fgColor;
		BOOL clipOn = !((clipboard.endRow == -1) || ((clipboard.endRow == clipboard.beginRow) &&
						  (clipboard.endColumn == clipboard.beginColumn)));
		COLORREF selectedBg = GetSysColor(COLOR_HIGHLIGHT);
		COLORREF selectedFg = GetSysColor(COLOR_HIGHLIGHTTEXT);
		char strOutput[2];
		int r, c;

		// What area needs to be repainted?
		GetUpdateRect(hWnd, &rect, FALSE);

		// Calculate update rectangle boundries in terms of character positions
		if ((rect.top = (rect.top / SCR_Y) - 1) < 0)
		  rect.top = 0;
		if ((rect.bottom = (rect.bottom / SCR_Y) + 1) > numRows - 1)
		  rect.bottom = numRows - 1;
		if ((rect.left = (rect.left / SCR_X) - 1) < 0)
		  rect.left = 0;
		if ((rect.right = (rect.top / SCR_Y) - 1) > numColumns - 1)
		  rect.right = numColumns - 1;

		// Set the default text and background colors
		SetTextColor(hDC, curColor.fgColor);
		SetBkColor(hDC, curColor.bgColor);

		// Initialize the strOutput variable
		strOutput[1] = '\0';

		// Disable the caret
		SetCaretState(FALSE, FALSE);

		// Repaint every character individually
		for (r = rect.top; r <= rect.bottom; r++)
		{
		  for (c = rect.left; c <= rect.right; c++)
		  {
			 // If there are any characters selected...
			 if (clipOn && isCharSelected(r, c))
			 {
				// Set the color to the "selected" color
				if (bgColor != selectedBg)
				{
				  bgColor = selectedBg;
				  SetBkColor(hDC, bgColor);
				}
				if (fgColor != selectedFg)
				{
				  fgColor = selectedFg;
				  SetTextColor(hDC, fgColor);
				}
			 }
			 else
			 {
				// Set the character's colors, if setting is required
				if (fgColor != color[r][c]->fgColor)
				{
				  fgColor = color[r][c]->fgColor;
				  SetTextColor(hDC, fgColor);
				}
				if (bgColor != color[r][c]->bgColor)
				{
				  bgColor = color[r][c]->bgColor;
				  SetBkColor(hDC, bgColor);
				}
			 }

			 // Write the character to the screen
			 strOutput[0] = scr[r][c];
			 TextOut(hDC, c * SCR_X, r * SCR_Y, strOutput, 1);
		  }
		}

		// Restore the DC and release it
		SelectPalette(hDC, hOldPal, FALSE);
		SelectObject(hDC, hOldFont);
		ReleaseDC(hWnd, hDC);
		ValidateRect(hWnd, NULL);

		// Enable the caret
		SetCaretState(TRUE, TRUE);
		return 0;
	 } /* WM_PAINT */

	 case WM_CLOSE :
	 {
		// The user chose to close the window; send a DestroyWindow
		DestroyWindow(hWnd);
		return 0;
	 }

	 case WM_DESTROY :
	 {
		// The window is being destroyed; remove us from the clipboard chain
		ChangeClipboardChain(hWnd, clipboard.hNext);

		// Clear the caret
		SetCaretState(FALSE, FALSE);

		// Post the WM_QUIT message so that the
		// message loop may exit, thus ending this program
		PostQuitMessage(0);
		return 0;
	 }

	 case WM_CHAR :
	 {
		// A character was typed in; place it into the input buffer
		if (lstrlen(scrIn) == scrInLength)
		{
		  // No more characters are available; play a MB_ICONEXCLAMATION sound
		  MessageBeep(MB_ICONEXCLAMATION);
		}
		else
		{
		  // There is room; add this character to the input buffer
		  UINT len = lstrlen(scrIn);
		  scrIn[len] = (char) wParam;
		  scrIn[len + 1] = '\0';

		  // If we are to echo the characters, display this on the screen
		  // and on the screen buffer
		  if ((echoChars) && (wParam != '\r'))
		  {
			 char scrOut[2];
			 scrOut[0] = (char) wParam;
			 scrOut[1] = '\0';
			 SendMessage(hWnd, SCR_PERFORMFUNCTION, SCR_PRINTF, (LPARAM) scrOut);
		  }
		}
		return 0;
	 } /* WM_CHAR */

	 case WM_COMMAND :
	 {
		// WM_COMMAND variable
		UINT idItem = wParam;

		switch (idItem)
		{
		  case CM_EXIT :
		  {
			 // Exit menu item selected; destroy window
			 DestroyWindow(hWnd);
			 break;
		  }

		  case CM_ABOUT :
		  {
			 // About selected; display About dialog box
			 DLGPROC dlgProc = (DLGPROC) MakeProcInstance((FARPROC) AboutDlgProc, hInstance);
			 DialogBox(hInstance, "ABOUT", hWnd, dlgProc);
			 FreeProcInstance((FARPROC) dlgProc);
			 break;
		  }

		  case CM_EDITPASTE :
		  {
			 // Paste the clipboard contents into the input buffer if it's okay
			 if (IsClipboardFormatAvailable(CF_TEXT))
			 {
				if (OpenClipboard(hWnd))
				{
				  // Put the clipboard data into the input buffer
				  HGLOBAL hstr = GetClipboardData(CF_TEXT);
				  LPSTR str = GlobalLock(hstr);
				  UINT i, slen = lstrlen(str);
				  UINT len, numNLs = 0;
				  i = len = lstrlen(scrIn);

				  // If there is enough room, then we can paste :-)
				  if (scrInLength - len + 1 >= lstrlen(str))
				  {
					 // Paste until we meet the end of the buffer, or the text;
					 // strip the newlines in the process
					 while ((i < scrInLength) && (i - len + numNLs < slen))
					 {
						if (str[i - len + numNLs] == '\n')
						{
						  // Strip the newline by just increasing numNLs
						  numNLs++;
						}
						else
						{
						  // Assign this value to the end of the input buffer
						  scrIn[i] = str[i - len + numNLs];
						  i++;
						}
					 }
					 scrIn[i] = '\0';

					 // If we are to echo the characters, display this on the screen
					 // and on the screen buffer
					 if (echoChars)
						SendMessage(hWnd, SCR_PERFORMFUNCTION, SCR_PRINTF, (LPARAM) str);
				  } /* if there is enough room... */

				  // Close the clipboard after unlocking the object
				  GlobalUnlock(hstr);
				  CloseClipboard();
				}
			 }
			 break;
		  } /* CM_EDITPASTE */

		  case CM_EDITSELECTALL :
		  {
			 // Select everything on the screen
			 clipboard.beginRow = 0;
			 clipboard.beginColumn = 0;
			 clipboard.endRow = numRows - 1;
			 clipboard.endColumn = numColumns - 1;

			 // Repaint the screen
			 InvalidateRect(hWnd, NULL, FALSE);
			 break;
		  }

		  case CM_EDITCOPY :
		  {
			 // Copy the selected text to the clipboard
			 HGLOBAL hstr;
			 LPSTR str;
			 UINT len = getSelectedTextLength();

			 // If there is text to copy, copy it
			 if (len > 0)
			 {
				if (OpenClipboard(hWnd))
				{
				  hstr = GlobalAlloc(GHND | GMEM_SHARE, sizeof(BYTE) * (len + 1));
				  str = GlobalLock(hstr);
				  getSelectedText(str);
				  EmptyClipboard();
				  SetClipboardData(CF_TEXT, hstr);
				  CloseClipboard();
				}
			 }
			 break;
		  } /* CM_EDITCOPY */

		  case CM_CLOSEONEXIT :
		  {
			 // Change the closeOnExit flag
			 closeOnExit = !closeOnExit;

			 // Set the closeOnExit menu item
			 if (closeOnExit)
				CheckMenuItem(hMenu, CM_CLOSEONEXIT, MF_BYCOMMAND | MF_CHECKED);
			 else
				CheckMenuItem(hMenu, CM_CLOSEONEXIT, MF_BYCOMMAND | MF_UNCHECKED);
			 break;
		  }

		} /* switch (idItem) */

		return 0;
	 } /* WM_COMMAND */
  } /* switch (message) */


  // FUNCTION PERFORMER
  if (message == SCR_PERFORMFUNCTION)
  {
	 // Determine the type of message
	 switch (wParam)
	 {
		case SCR_PRINTF :
		{
		  // Get the DC and initialize the DC characteristics with the font and palette
		  LPSTR str = (LPSTR) lParam;
		  HDC hDC = GetDC(hWnd);
		  HFONT hOldFont = SelectObject(hDC, hFont);
		  HPALETTE hOldPal = SelectPalette(hDC, hPalette, FALSE);
		  COLORREF selectedBg = GetSysColor(COLOR_HIGHLIGHT);
		  COLORREF selectedFg = GetSysColor(COLOR_HIGHLIGHTTEXT);
		  int i, end;

		  // Hide the caret
		  SetCaretState(FALSE, FALSE);

		  // Set the text and background colors
		  SetTextColor(hDC, curColor.fgColor);
		  SetBkColor(hDC, curColor.bgColor);

		  // Process every byte until the terminating NULL is read, or the EOLN
		  i = 0;
		  end = lstrlen(str) - 1;
		  while (i <= end)
		  {
			 switch (str[i])
			 {
				case '\n' :
				{
				  // Line Feed
				  curColumn = 0;
				  curRow++;
				  if (curRow > numRows - 1)
				  {
					 // Scroll the current window up by one
					 LPSTR firstRow = scr[0];
					 LPCOLORINFO *colorRow = color[0];
					 RECT rect, newRect;
					 int r;

					 // Initialize rectangle variables
					 rect.top = 0;
					 rect.bottom = numRows * SCR_Y;
					 rect.left = 0;
					 rect.right = numColumns * SCR_X;

					 // Scroll the DC
					 ScrollDC(hDC, 0, -1 * SCR_Y, &rect, NULL, NULL, &newRect);

					 // Reflect this change in memory by moving every row up by one
					 for (r = 0; r <= numRows - 2; r++)
					 {
						scr[r] = scr[r + 1];
						color[r] = color[r + 1];
					 }
					 scr[numRows - 1] = firstRow;
					 color[numRows - 1] = colorRow;

					 // Clear the bottom row on the screen
					 memset(scr[numRows - 1], ' ', numColumns);
					 TextOut(hDC, 0, (numRows - 1) * SCR_Y, scr[numRows - 1], numColumns);

					 // Clear the bottom row in memory (with NULLs this time)
					 memset(scr[numRows - 1], '\0', numColumns);

					 // Set every color in the bottom row to the current colors
					 for (r = 0; r <= numColumns - 1; r++)
					 {
						color[numRows - 1][r]->bgColor = curColor.bgColor;
						color[numRows - 1][r]->fgColor = curColor.fgColor;
					 }

					 // Set the current row to the last one
					 curRow--;
				  }
				  break;
				}

				case '\r' :
				{
				  // Enter Key (carriage return, NOT line feed)
				  curColumn = 0;
				  break;
				}

				case '\b' :
				{
				  // Backspace

				  // If we are at the input starting place, then go up
				  if (!((curRow == inRow) && (curColumn == inColumn)))
				  {
					 if ((int) curColumn - 1 >= 0)
					 {
						// Erase the character before us and reflect that in the buffer
						TextOut(hDC, (curColumn - 1) * SCR_X, curRow * SCR_Y, " ", 1);
						curColumn--;
						scr[curRow][curColumn] = ' ';
						color[curRow][curColumn]->bgColor = curColor.bgColor;
						color[curRow][curColumn]->fgColor = curColor.fgColor;
					 }
					 else
					 {
						if ((int) curRow - 1 >= 0)
						{
						  // Go to the end of the previous line and delete that character
						  curRow--;

						  // If the previous line has anything in it...
						  if (getLastCharPos(scr[curRow]) >= 0)
						  {
							 // Erase the character before us and reflect that in the buffer
							 curColumn = getLastCharPos(scr[curRow]);
							 TextOut(hDC, curColumn * SCR_X, curRow * SCR_Y, " ", 1);
							 scr[curRow][curColumn] = ' ';
							 color[curRow][curColumn]->bgColor = curColor.bgColor;
							 color[curRow][curColumn]->fgColor = curColor.fgColor;
						  }
						  else
							 curColumn = 0;
						}
					 }
				  }
				  break;
				}

				case '\t' :
				{
				  // Tab; display 1 space * tabSpace
				  UINT i;
				  for (i = 0; i <= tabSpace - 1; i++)
				  {
					 TextOut(hDC, (curColumn + i) * SCR_X, curRow * SCR_Y, " ", 1);
					 scr[curRow][curColumn + i] = ' ';
				  }
				  curColumn += tabSpace;
				  break;
				}

				default :
				{
				  // A printable character

				  // Be a psychic and look into the future to see how many printable
				  // characters we can handle at the same time in the same line :-)
				  LPSTR strOutput;
				  HGLOBAL hstrOutput;
				  int future = i + 1;
				  int g;
				  BOOL selected = isCharSelected(curRow, curColumn);

				  while (future <= end)
				  {
					 if ((str[future] == 10) || (str[future] == 13) || (str[future] == 8) || (str[future] == 9) || ((future - i) + curColumn > numColumns - 1) || (isCharSelected(curRow, curColumn + future - i) != selected))
						break;
					 future++;
				  }
				  future--;

				  // Allocate and assign the strOutput string
				  hstrOutput = GlobalAlloc(GHND, sizeof(BYTE) * (future - i + 2));
				  strOutput = (LPSTR) GlobalLock(hstrOutput);
				  for (g = 0; g <= future - i; g++)
					 strOutput[g] = str[g + i];
				  strOutput[future - i + 1] = '\0';

				  // Set the selection color (if selection persists)
				  if (selected)
				  {
					 if (curColor.bgColor != selectedBg)
						SetBkColor(hDC, selectedBg);
					 if (curColor.fgColor != selectedFg)
						SetTextColor(hDC, selectedFg);
				  }

				  // Write the string
				  TextOut(hDC, curColumn * SCR_X, curRow * SCR_Y, strOutput, lstrlen(strOutput));

				  // Restore the normal color (if selection persists)
				  if (selected)
				  {
					 SetBkColor(hDC, curColor.bgColor);
					 SetTextColor(hDC, curColor.fgColor);
				  }

				  // Reflect these changes in the buffer
				  for (g = curColumn; g <= curColumn + (future - i); g++)
				  {
					 scr[curRow][g] = str[(g - curColumn) + i];
					 color[curRow][g]->bgColor = curColor.bgColor;
					 color[curRow][g]->fgColor = curColor.fgColor;
				  }

				  // Advance the row and column positions
				  curColumn += (future - i) + 1;
				  i = future;
				  if (curColumn == numColumns)
				  {
					 // It is time to go to the next line
					 curColumn = 0;
					 curRow++;

					 // If our row is now after the end, set the current character
					 // to a newline so that appropriate processing may take place
					 if (curRow > numRows - 1)
					 {
						curRow--;
						i = future - 1;
						str[future] = '\n';
					 }
				  }

				  // Free the memory object
				  GlobalUnlock(hstrOutput);
				  GlobalFree(hstrOutput);
				  break;
				}

			 } /* switch */

			 // Process the next character in the output buffer
			 i++;

		  } /* while */

		  // Done with processing; free str and "clean up" the DC
		  SelectPalette(hDC, hOldPal, FALSE);
		  SelectObject(hDC, hOldFont);
		  ReleaseDC(hWnd, hDC);

		  // Set the caret position to the current row/column
		  SetCaretState(TRUE, TRUE);

		  // Return the new screen coordinates
		  return MAKELONG((WORD) curRow, (WORD) curColumn);
		} /* SCR_PRINTF */

		case SCR_SETPOS :
		{
		  // Hide the caret
		  SetCaretState(FALSE, FALSE);

		  // Set current position
		  if (LOWORD(lParam) < numRows)
			 curRow = LOWORD(lParam);
		  if (HIWORD(lParam) < numColumns)
			 curColumn = HIWORD(lParam);

		  // Set the caret position to the current row/column
		  SetCaretState(TRUE, TRUE);
		  return MAKELONG(curRow, curColumn);
		}

		case SCR_GETPOS :
		{
		  // Get the current position
		  return MAKELONG(curRow, curColumn);
		}

		case SCR_KBHIT :
		{
		  // Has a key been pressed (or hit)?  Return the number of bytes if there are some
		  return lstrlen(scrIn);
		}

		case SCR_SETCLOSE :
		{
		  // Set the new value of the close flag
		  closeOnExit = (BOOL) lParam;
		  return lParam;
		}

		case SCR_GETCLOSE :
		{
		  // Get the current closeOnExit flag
		  return closeOnExit;
		}

		case SCR_SETINPUTLEN :
		{
		  // Set the new input buffer length by freeing the old and creating the new
		  HGLOBAL newBuffer;
		  GlobalUnlock(hscrIn);

		  // If the allocation routine did not work, return FALSE
		  if ((newBuffer = GlobalReAlloc(hscrIn, sizeof(BYTE) * (lParam + 1), NULL)) == NULL)
			 return FALSE;

		  // Set the new handle and lock it
		  hscrIn = newBuffer;
		  scrIn = GlobalLock(hscrIn);
		  scrInLength = (UINT) lParam + 1;

		  // Reset the new buffer and return TRUE
		  scrIn[0] = '\0';
		  return TRUE;
		}

		case SCR_GETINPUTLEN :
		{
		  // Get the length of the current input buffer
		  return (scrInLength - 1);
		}

		case SCR_GETEXTENTS :
		{
		  // Get the current screen extents (numRows and numColumns)
		  return MAKELONG(numRows, numColumns);
		}

		case SCR_DISPLAYABOUT :
		{
		  // Raise the About dialog box by "selecting it from the menu"
		  SendMessage(hWnd, WM_COMMAND, CM_ABOUT, NULL);
		  return NULL;
		}

		case SCR_ABOUTTEXT :
		{
		  // Set the about dialog box text
		  HGLOBAL newBuffer;
		  LPSTR aboutText;

		  // If the allocation routine did not work, return FALSE
		  if ((newBuffer = GlobalReAlloc(haboutText, sizeof(BYTE) * (lstrlen((LPSTR) lParam) + 1), NULL)) == NULL)
			 return FALSE;

		  // Set the new handle and set its value
		  haboutText = newBuffer;
		  aboutText = GlobalLock(haboutText);
		  lstrcpy(aboutText, (LPSTR) lParam);
		  GlobalUnlock(haboutText);
		  return TRUE;
		}

		case SCR_GETFONTMULTIPLIERS :
		{
		  // Return the font multipliers
		  return MAKELONG(SCR_X, SCR_Y);
		}

		case SCR_SCANF :
		{
		  // Wait until the user presses [Enter], or had pressed it
		  LPSTR str = (LPSTR) lParam;
		  int end, i;
		  int pos;

		  // Get an input
		  inRow = curRow;
		  inColumn = curColumn;
		  echoChars = TRUE;
		  waitForEoln();
		  echoChars = FALSE;
		  inRow = 0;
		  inColumn = 0;

		  // Find out where the carriage return is
		  pos = 0;
		  end = lstrlen(scrIn) - 1;
		  while ((scrIn[pos] != '\r') && (pos <= end))
			 pos++;

		  // Something different is happening; either we're exiting or a
		  // function didn't work; return FALSE
		  if (pos > end)
			 return FALSE;

		  // Copy our part to the user's string, minus the carriage return
		  lstrcpyn(str, scrIn, pos + 1);

		  // Move the other characters in the input buffer down (including terminating NULL)
		  end++;
		  pos++;
		  for (i = pos; i <= end; i++)
			 scrIn[i - pos] = scrIn[i];
		  return TRUE;
		} /* SCR_SCANF */

		case SCR_READBYTE :
		{
		  // Wait until the user presses [Enter], or had pressed it
		  BYTE byteToRead;
		  UINT i;
		  UINT len = lstrlen(scrIn);
		  BOOL eolnFlag = (BOOL) lParam;

		  if (eolnFlag)
		  {
			 // Read the entire line first
			 echoChars = TRUE;
			 inRow = curRow;
			 inColumn = curColumn;
			 waitForEoln();
			 echoChars = FALSE;
			 inRow = 0;
			 inColumn = 0;
			 len = lstrlen(scrIn);
		  }
		  else
		  {
			 // Secure a byte
			 if (len == 0)
			 {
				if (!waitForChar())
				  return NULL;
				len = lstrlen(scrIn);
			 }
		  }

		  // Assign byteToRead to the next byte
		  byteToRead = scrIn[0];

		  // Move the other characters in the input buffer down (including terminating NULL)
		  for (i = 1; i <= len; i++)
			 scrIn[i - 1] = scrIn[i];

		  // Return the byte
		  return byteToRead;
		} /* SCR_READBYTE */

		case SCR_SETFGCOLOR :
		{
		  // Set the foreground color
		  curColor.fgColor = (COLORREF) lParam;
		  return lParam;
		}

		case SCR_SETBGCOLOR :
		{
		  // Set the background color
		  curColor.bgColor = (COLORREF) lParam;
		  return lParam;
		}

		case SCR_GETFGCOLOR :
		{
		  // Get the foreground color
		  return (LRESULT) curColor.fgColor;
		}

		case SCR_GETBGCOLOR :
		{
		  // Get the background color
		  return (LRESULT) curColor.bgColor;
		}

		case SCR_CLEARINPUT :
		{
		  // Clear the input buffer
		  scrIn[0] = '\0';
		  return NULL;
		}

		case SCR_CLEARLINE :
		{
		  // Clear the text on the current row from the current column to the last column
		  HDC hDC = GetDC(hWnd);
		  HFONT hOldFont = SelectObject(hDC, hFont);
		  HPALETTE hOldPal = SelectPalette(hDC, hPalette, FALSE);
		  UINT i;

		  // Set the background color and write the spaces while reflecting in memory
		  SetBkColor(hDC, curColor.bgColor);
		  for (i = curColumn; i <= numColumns - 1; i++)
		  {
			 scr[curRow][i] = ' ';
			 color[curRow][i]->bgColor = curColor.bgColor;
			 color[curRow][i]->fgColor = curColor.fgColor;
			 TextOut(hDC, i * SCR_X, curRow * SCR_Y, " ", 1);
		  }

		  // Clean up the device context and return NULL
		  SelectPalette(hDC, hOldPal, FALSE);
		  SelectObject(hDC, hOldFont);
		  ReleaseDC(hWnd, hDC);
		  return NULL;
		} /* SCR_CLEARLINE */

		case SCR_APPCLOSE :
		{
		  // Close the window by calling CM_EXIT, if we close on exit
		  if (closeOnExit)
			 SendMessage(hWnd, WM_COMMAND, CM_EXIT, NULL);
		  return NULL;
		}

		case SCR_CANPASTE :
		{
		  // Return whether or not the current clipboard data can paste
		  return IsClipboardFormatAvailable(CF_TEXT);
		}

		case SCR_PASTE :
		{
		  // Paste by calling that function through the window
		  SendMessage(hWnd, WM_COMMAND, CM_EDITPASTE, NULL);
		  return NULL;
		}

		case SCR_GETSELECTION :
		{
		  // Get the current selection start/end variables
		  return MAKELONG(MAKEWORD((BYTE) clipboard.beginRow, (BYTE) clipboard.beginColumn),
								MAKEWORD((BYTE) clipboard.endRow, (BYTE) clipboard.endColumn));
		}

		case SCR_SETSELECTION :
		{
		  // Perform parameter checks (return FALSE if there is an error)
		  int beginRow, beginColumn, endRow, endColumn;
		  beginRow = LOWORD(LOBYTE(lParam));
		  beginColumn = LOWORD(HIBYTE(lParam));
		  endRow = HIWORD(LOBYTE(lParam));
		  endColumn = HIWORD(HIBYTE(lParam));
		  if ((beginRow < 0) || (beginRow > numRows - 1))
			 return FALSE;
		  if ((beginColumn < 0) || (beginColumn > numColumns - 1))
			 return FALSE;
		  if ((endRow < 0) || (endRow > numRows - 1))
			 return FALSE;
		  if ((endColumn < 0) || (endColumn > numColumns - 1))
			 return FALSE;
		  if (endRow < beginRow)
			 return FALSE;

		  // Set the current selection
		  clipboard.beginRow = beginRow;
		  clipboard.beginColumn = beginColumn;
		  clipboard.endRow = endRow;
		  clipboard.endColumn = endColumn;

		  // Reflect this change on the screen
		  InvalidateRect(hWnd, NULL, FALSE);
		  SendMessage(hWnd, WM_PAINT, NULL, NULL);
		  return lParam;
		} /* SCR_SETSELECTION */

		case SCR_SELECTALL :
		{
		  // Select everything
		  SendMessage(hWnd, WM_COMMAND, CM_EDITSELECTALL, NULL);
		  return NULL;
		}

		case SCR_GETSELLENGTH :
		{
		  // Get the length of the current selection
		  return getSelectedTextLength();
		}

		case SCR_GETSELTEXT :
		{
		  // Copy the selected text into str (program is assumed to have alloted enough
		  // memory with SCR_GETSELLENGTH)
		  LPSTR str = (LPSTR) lParam;
		  return getSelectedText(str);
		}

		case SCR_COPY :
		{
		  // Copy using CM_EDITCOPY
		  SendMessage(hWnd, WM_COMMAND, CM_EDITCOPY, NULL);
		  return NULL;
		}

		case SCR_CLEARSCR :
		{
		  // Clear the entire screen!!
		  HDC hDC = GetDC(hWnd);
		  HFONT hOldFont = SelectObject(hDC, hFont);
		  HPALETTE hOldPal = SelectPalette(hDC, hPalette, FALSE);
		  HBRUSH hBrush = CreateSolidBrush(curColor.bgColor);
		  HBRUSH hOldBrush = SelectObject(hDC, hBrush);
		  RECT rect;
		  UINT r, c;

		  // Hide the caret
		  SetCaretState(FALSE, FALSE);

		  // Clear the memory
		  for (r = 0; r < numRows; r++)
		  {
			 for (c = 0; c < numColumns; c++)
			 {
				scr[r][c] = ' ';
				color[r][c]->fgColor = curColor.fgColor;
				color[r][c]->bgColor = curColor.bgColor;
			 }
		  }
		  curRow = 0;
		  curColumn = 0;

		  // Set the background color and write a large rectangle
		  SetBkColor(hDC, curColor.bgColor);
		  rect.top = 0;
		  rect.left = 0;
		  rect.right = numColumns * SCR_X + 1;
		  rect.bottom = numRows * SCR_Y + 1;
		  FillRect(hDC, &rect, hBrush);

		  // Clean up the device context and return NULL
		  SelectObject(hDC, hOldBrush);
		  SelectPalette(hDC, hOldPal, FALSE);
		  SelectObject(hDC, hOldFont);
		  ReleaseDC(hWnd, hDC);

		  // Show the caret
		  SetCaretState(TRUE, TRUE);
		  return NULL;
		}

	 } /* switch (wParam) */
  } /* if SCR_PERFORMFUNCTION */
  return DefWindowProc(hWnd, message, wParam, lParam);
}

HPALETTE makePalette(void)
{
  HPALETTE hPal;
  HGLOBAL hGlobal;
  LPLOGPALETTE lp;

  // Create the palette with the standard colors (including brown)
  hGlobal = GlobalAlloc(GHND, sizeof(LOGPALETTE) + 17 * sizeof(PALETTEENTRY));
  lp = (LPLOGPALETTE) GlobalLock(hGlobal);
  lp->palVersion = 0x300;
  lp->palNumEntries = 17;
  lp->palPalEntry[0].peRed = 0;
  lp->palPalEntry[0].peGreen = 0;
  lp->palPalEntry[0].peBlue = 0;
  lp->palPalEntry[0].peFlags = NULL;
  lp->palPalEntry[1].peRed = 191;
  lp->palPalEntry[1].peGreen = 0;
  lp->palPalEntry[1].peBlue = 0;
  lp->palPalEntry[1].peFlags = NULL;
  lp->palPalEntry[2].peRed = 0;
  lp->palPalEntry[2].peGreen = 191;
  lp->palPalEntry[2].peBlue = 0;
  lp->palPalEntry[2].peFlags = NULL;
  lp->palPalEntry[3].peRed = 128;
  lp->palPalEntry[3].peGreen = 128;
  lp->palPalEntry[3].peBlue = 0;
  lp->palPalEntry[3].peFlags = NULL;
  lp->palPalEntry[4].peRed = 0;
  lp->palPalEntry[4].peGreen = 0;
  lp->palPalEntry[4].peBlue = 191;
  lp->palPalEntry[4].peFlags = NULL;
  lp->palPalEntry[5].peRed = 191;
  lp->palPalEntry[5].peGreen = 0;
  lp->palPalEntry[5].peBlue = 191;
  lp->palPalEntry[5].peFlags = NULL;
  lp->palPalEntry[6].peRed = 0;
  lp->palPalEntry[6].peGreen = 191;
  lp->palPalEntry[6].peBlue = 191;
  lp->palPalEntry[6].peFlags = NULL;
  lp->palPalEntry[7].peRed = 192;
  lp->palPalEntry[7].peGreen = 192;
  lp->palPalEntry[7].peBlue = 192;
  lp->palPalEntry[7].peFlags = NULL;
  lp->palPalEntry[8].peRed = 128;
  lp->palPalEntry[8].peGreen = 128;
  lp->palPalEntry[8].peBlue = 128;
  lp->palPalEntry[8].peFlags = NULL;
  lp->palPalEntry[9].peRed = 255;
  lp->palPalEntry[9].peGreen = 0;
  lp->palPalEntry[9].peBlue = 0;
  lp->palPalEntry[9].peFlags = NULL;
  lp->palPalEntry[10].peRed = 0;
  lp->palPalEntry[10].peGreen = 255;
  lp->palPalEntry[10].peBlue = 0;
  lp->palPalEntry[10].peFlags = NULL;
  lp->palPalEntry[11].peRed = 255;
  lp->palPalEntry[11].peGreen = 255;
  lp->palPalEntry[11].peBlue = 0;
  lp->palPalEntry[11].peFlags = NULL;
  lp->palPalEntry[12].peRed = 0;
  lp->palPalEntry[12].peGreen = 0;
  lp->palPalEntry[12].peBlue = 255;
  lp->palPalEntry[12].peFlags = NULL;
  lp->palPalEntry[13].peRed = 255;
  lp->palPalEntry[13].peGreen = 0;
  lp->palPalEntry[13].peBlue = 255;
  lp->palPalEntry[13].peFlags = NULL;
  lp->palPalEntry[14].peRed = 0;
  lp->palPalEntry[14].peGreen = 255;
  lp->palPalEntry[14].peBlue = 255;
  lp->palPalEntry[14].peFlags = NULL;
  lp->palPalEntry[15].peRed = 255;
  lp->palPalEntry[15].peGreen = 255;
  lp->palPalEntry[15].peBlue = 255;
  lp->palPalEntry[15].peFlags = NULL;
  lp->palPalEntry[16].peRed = 170;
  lp->palPalEntry[16].peGreen = 85;
  lp->palPalEntry[16].peBlue = 0;
  lp->palPalEntry[16].peFlags = NULL;
  hPal = CreatePalette(lp);
  GlobalUnlock(hGlobal);
  GlobalFree(hGlobal);

  // Return the result of this hard work :-)
  return hPal;
}

// About dialog box procedure
#pragma argsused
BOOL CALLBACK _export AboutDlgProc(HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
  switch (msg)
  {
	 case WM_INITDIALOG :
	 {
		// Set about text
		LPSTR aboutText = GlobalLock(haboutText);
		SetDlgItemText(hDlg, IDC_ABOUTTEXT, aboutText);
		GlobalUnlock(haboutText);
		return TRUE;
	 }

	 case WM_COMMAND :
	 {
		// If OK or [Esc] was pressed, exit this dialog
		if ((wParam == IDOK) || (wParam == IDCANCEL))
		  EndDialog(hDlg, 0);
		return 0;
	 }
  }
  return FALSE;
}

UINT getSelectedTextLength(void)
{
  UINT len;

  // If this selection spans more than 1 line, then calculate accordingly
  if (clipboard.endRow > clipboard.beginRow)
  {
	 UINT i;

	 // Calculate the starting row
	 if ((getLastCharPos(scr[clipboard.beginRow]) - clipboard.beginColumn) < 0)
		len = 2;
	 else
		len = (getLastCharPos(scr[clipboard.beginRow]) - clipboard.beginColumn) + 3;

	 // Calculate the mid-section
	 for (i = clipboard.beginRow + 1; i <= clipboard.endRow - 1; i++)
		len += getLastCharPos(scr[i]) + 3;

	 // Calculate ending
	 len += (clipboard.endColumn <= getLastCharPos(scr[clipboard.endRow]))? clipboard.endColumn: getLastCharPos(scr[clipboard.endRow]) + 1;

	 // Return our results
	 return len;
  }
  else
  {
	 // If both equal -1, then return 0
	 if ((clipboard.beginRow == -1) || (clipboard.endRow == -1))
		return 0;

	 // The string is within the text of this row; return the end - the begin + 1
	 return clipboard.endColumn - clipboard.beginColumn + 1;
  }
}

BOOL getSelectedText(LPSTR str)
{
  // Copy the currently selected text into str
  int i, stri;

  // If nothing's selected, return FALSE
  if ((clipboard.beginRow == -1) || (clipboard.endRow == -1))
	 return FALSE;

  stri = 0;
  if (clipboard.endRow > clipboard.beginRow)
  {
	 // Get the starting row
	 int t, pos;
	 pos = getLastCharPos(scr[clipboard.beginRow]);
	 for (i = clipboard.beginColumn; i <= pos; i++)
	 {
		str[stri] = scr[clipboard.beginRow][i];
		stri++;
	 }
	 str[stri] = '\r';
	 str[stri + 1] = '\n';
	 stri += 2;

	 // Get the "in-between" rows, if any
	 for (i = clipboard.beginRow + 1; i <= clipboard.endRow - 1; i++)
	 {
		pos = getLastCharPos(scr[i]);
		for (t = 0; t <= pos; t++)
		{
		  str[stri] = scr[i][t];
		  stri++;
		}
		str[stri] = '\r';
		str[stri + 1] = '\n';
		stri += 2;
	 }

	 // Get the last row
	 pos = (clipboard.endColumn <= getLastCharPos(scr[clipboard.endRow]))? clipboard.endColumn: getLastCharPos(scr[clipboard.endRow]);
	 for (i = 0; i <= pos; i++)
	 {
		str[stri] = scr[clipboard.endRow][i];
		stri++;
	 }
	 str[stri] = '\0';
	 return TRUE;
  }
  else
  {
	 // Read from start to finish on the same row
	 for (i = clipboard.beginColumn; i <= clipboard.endColumn; i++)
	 {
		str[stri] = scr[clipboard.beginRow][i];
		stri++;
	 }
	 str[stri] = '\0';
	 return TRUE;
  }
}

void SetCaretState(BOOL state, BOOL on)
{
  // Change the status of the caret
  if (on != caretActive)
  {
	 if (on)
	 {
		// Enable the caret
		SetTimer(hWnd, CARETID, GetCaretBlinkTime(), NULL);
		caretActive = TRUE;
		state = TRUE;
	 }
	 else
	 {
		// Disable the caret
		KillTimer(hWnd, CARETID);
		caretActive = FALSE;
		state = FALSE;
	 }
  }

  if ((caretStatus != state) && !caretHide)
  {
	 // Invert the area of the caret
	 HDC hDC = GetDC(hWnd);
	 RECT rect;
	 rect.left = curColumn * SCR_X;
	 rect.top = ((curRow + 1) * SCR_Y) - SCR_Y / 8;
	 rect.bottom = ((curRow + 1) * SCR_Y);
	 rect.right = (curColumn + 1) * SCR_X;

	 InvertRect(hDC, &rect);
	 caretStatus = state;
	 ReleaseDC(hWnd, hDC);
  }
}

void waitForEoln(void)
{
  // Wait until there '\r' exists in the input buffer

  // Display any text in the buffer up to the 'r', if any
  int i, end, t;
  BOOL nextLine = FALSE;
  if (strstr(scrIn, "\r") != NULL)
  {
	 char scrOut[2];
	 i = 0;
	 while (scrIn[i] != '\r')
	 {
		scrOut[0] = scrIn[i];
		scrOut[1] = '\0';
		SendMessage(hWnd, SCR_PERFORMFUNCTION, SCR_PRINTF, (LPARAM) scrOut);
		i++;
	 }
  }

  // Remove all the backspaces that exist
  i = 0;
  end = lstrlen(scrIn) - 1;
  while ((i <= end) && (i >= 0))
  {
	 // If we encounter a carriage return, stop removing backspaces
	 if (scrIn[i] == '\r')
		break;

	 // Backspace encountered; process it!
	 if (scrIn[i] == '\b')
	 {
		if (i > 0)
		{
		  for (t = i; t <= end; t++)
			 scrIn[t - 1] = scrIn[t + 1];
		  end -= 2;
		  i--;
		}
		else
		{
		  for (t = i; t <= end; t++)
			 scrIn[t] = scrIn[t + 1];
		  end--;
		}
	 }
	 else
		i++;
  }

  // Loop until we find a carriage return
  while (strstr(scrIn, "\r") == NULL)
  {
	 // Wait for the next character
	 if (!waitForChar())
		return;

	 // Check next character; if it's a backspace, process it
	 i = (int) lstrlen(scrIn) - 1;
	 if (i >= 0)
	 {
		if (scrIn[i] == '\b')
		{
		  if (i > 0)
			 i--;
		  scrIn[i] = '\0';
		}
	 }
	 nextLine = TRUE;
  }

  if (nextLine)
  {
	 // Advance to the next line if the [Enter] was pressed
	 SendMessage(hWnd, SCR_PERFORMFUNCTION, SCR_PRINTF, (LPARAM) "\n");
  }
}

BOOL waitForChar(void)
{
  // Message Loop that quits with WM_QUIT or when a character was just typed
  MSG msg;
  while (TRUE)
  {
	 // Wait for another message and get that message
	 WaitMessage();
	 PeekMessage(&msg, NULL, 0, 0, PM_REMOVE);

	 // Standard message handling
	 TranslateMessage(&msg);
	 TranslateAccelerator(hWnd, hAcc, &msg);
	 DispatchMessage(&msg);

	 // User is quitting; exit this procedure!!
	 if (msg.message == WM_QUIT)
		return FALSE;

	 // A character was just pressed!
	 if (msg.message == WM_CHAR)
		return TRUE;
  }
}

int getLastCharPos(LPCSTR line)
{
  int pos = numColumns - 1;
  while (pos >= 0)
  {
	 if (line[pos] != ' ')
		break;
	 pos--;
  }
  return pos;
}

BOOL isCharSelected(UINT r, UINT c)
{
  // Determine if the current character is selected
  if ((clipboard.endRow == -1) && (clipboard.beginRow == -1))
	 return FALSE;
  if ((r <= clipboard.endRow) && (r >= clipboard.beginRow))
  {
	 if (r == clipboard.beginRow)
	 {
		if (r == clipboard.endRow)
		  return (c >= clipboard.beginColumn) && (c <= clipboard.endColumn);
		else
		  return (c >= clipboard.beginColumn);
	 }
	 else
	 {
		if (r == clipboard.endRow)
		  return (c <= clipboard.endColumn);
		else
		  return TRUE;
	 }
  }
  else
	 return FALSE;
}

